#!/usr/bin/env python
#Copyright (c) 2009 Jeff Bryner
#http://jeffbryner.com
#python script to gather googletoolbar traffic and snoop on it.
#This program is free software; you can redistribute it and/or modify it under
#the terms of the GNU General Public License as published by the Free Software
#Foundation; either version 2 of the License, or (at your option) any later
#version.

#This program is distributed in the hope that it will be useful, but WITHOUT
#ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

#You should have received a copy of the GNU General Public License along with
#this program; if not, write to the Free Software Foundation, Inc.,
#59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import sys
import os
import re
import string
import fileinput
import getopt
import time
import base64
import binascii
import scapy 
from scapy.all import * 
from scapy.packet import * 
import nids

program= 'gtoolbarsnoop by jeff bryner'
version='v1.0'
safestringre=re.compile('[\x00-\x1F\x80-\xFF]')
def safestring(badstring):
        """makes a good strings out of a potentially bad one by escaping chars out of printable range"""
        return safestringre.sub('',badstring)

#regexes
faviconre=re.compile(r"""GET /favicon""",re.IGNORECASE)
useragentre=re.compile(r"""User-Agent:.*GoogleToolbar""",re.IGNORECASE|re.DOTALL)
bookmarksre=re.compile(r"""<bookmarks>""",re.IGNORECASE|re.DOTALL)
googlecookiere=re.compile(r"""Set-Cookie:.*domain=\.google\.com""",re.IGNORECASE|re.DOTALL)
titlere=re.compile(r"""<bookmark><title>(.*?)</title>""",re.IGNORECASE|re.DOTALL)
sload=""

end_states = (nids.NIDS_CLOSE, nids.NIDS_TIMEOUT, nids.NIDS_RESET, nids.NIDS_EXITING)




def handleTcpStream(tcp):
	if options['debug']:	
		print "tcps -", str(tcp.addr), " state:", tcp.nids_state
	if tcp.nids_state == nids.NIDS_JUST_EST:
        	# new to us, but is it one we're monitoring?
	        ((src, sport), (dst, dport)) = tcp.addr
        	if dport ==80 or sport==80:
			if options['debug']:	
		            print "collecting data between: " , tcp.addr
			tcp.client.collect = 1
			tcp.server.collect = 1
	elif tcp.nids_state == nids.NIDS_DATA:
	        # keep all of the stream's new data
	    	tcp.discard(0)

	#got anything to examine?
	if len(str(tcp.server.data[:tcp.server.count]).strip())>0 or tcp.nids_state in end_states:
		sload=""
	        ((src, sport), (dst, dport)) = tcp.addr	
		if options['debug']:
		        print "processing stream between:", tcp.addr		
		sload=str(tcp.server.data[:tcp.server.count]).strip()
		sload=sload + str(tcp.client.data[:tcp.client.count]).strip()

		if useragentre.findall(sload) and faviconre.findall(sload) and options['icons']:
			sys.stderr.write('gtoolbarsnoop: found a favicon hit\n')
			sys.stderr.write('Conversation: ' + str(tcp.addr) + '\n')
			sys.stderr.write('clientIP: ' + str(src) + '\n')			
			sys.stderr.write(str(tcp.server.data[:tcp.server.count]) + '\n')

		if bookmarksre.findall(sload) and options['bookmarks']:
			sys.stderr.write('gtoolbarsnoop: found bookmarks\n')
			sys.stderr.write('Conversation: ' + str(tcp.addr) + '\n')
			sys.stderr.write('clientIP: ' + str(src) + '\n')
			sys.stderr.write(str(tcp.client.data[:tcp.client.count]) + '\n')
			
		if titlere.findall(sload) and options['titles']:
			sys.stderr.write('gtoolbarsnoop: found titles\n')
			sys.stderr.write('clientIP: ' + str(src) + '\n')			
			for title in titlere.finditer(sload):
				sys.stderr.write(title.group(1) + '\n')

#debug
#	        print "addr:", tcp.addr
#	        print "To server:"
#	        print tcp.server.data[:tcp.server.count] # WARNING - may be binary
#	        print "To client:"
#	        print tcp.client.data[:tcp.client.count] # WARNING - as above


def main():
	global options
	options = read_options()
	nids.param("scan_num_hosts", 0)         # disable portscan detection

	if options['file'] <> '-':
		nids.param("filename", options['file'])
	nids.init()
	nids.register_tcp(handleTcpStream)
	# Loop forever (network device), or until EOF (pcap file)
	# Note that an exception in the callback will break the loop!
	try:
		nids.run()
	except nids.error, e:
		print "nids/pcap error:", e
	except Exception, e:
		print "misc. exception (runtime error in user callback?):", e
	sys.exit()
	

def read_options():
	"""Read options from config files and the command line, returns the 
	defaults if no user options are found"""

	# required options
	required = ['file']

	# defaults
	options = {'file'   : '',
        	   'debug'    : False,
        	   'verbose': False,
		   'icons': False,
		   'bookmarks': False,
		   'titles': False
		   }

    # read from command line
	helpstr = 'Usage: ' + sys.argv[0] + ' [OPTIONS]' + """\n
Options:
 * -f, --file   <filename>  the pcap input file to use
   -b, --bookmarks	    show bookmarks found
   -i, --icons	    	    show favicon hits
   -t, --titles		    show titles found
   -a, --all		    show all bookmarks, favicon hits and titles
   -d, --debug              run in debug mode with crazy output
   -h, --help   
   -v, --verbose            show some internal details
   -V, --version            show version info\n
A '*' means this option is required and has to be set."""

	optlist, args = getopt.getopt(sys.argv[1:], 'f:ibtahVv', ['file=', 'help', 'version', 'verbose' ,'debug','bookmarks','icons','titles','all'])

	# parse options
	for o, a in optlist:
		if (o == '-h' or o == '--help'):
			print helpstr
			sys.exit()
		elif (o == '-V' or o == '--version'):
			print program, version
			sys.exit()
		elif (o == '-v' or o == '--verbose'):
			options['verbose'] = True
		elif (o == '-d' or o == '--debug'):
			options['debug'] = True
		elif (o =='-i' or o == '--icons'):
			options['icons'] = True
		elif (o =='-b' or o == '--bookmarks'):			
			options['bookmarks'] = True
		elif (o =='-t' or o == '--titles'):			
			options['titles'] = True				
		elif (o =='-a' or o == '--all'):
			options['icons'] = True
			options['bookmarks'] = True
			options['titles'] = True			
		
		else:
			for option in options.keys():
				execcode = "if (o == '-%s' or o == '--%s'): options['%s'] = a" % (option[0], option, option)
				exec execcode

	for option in required:
		if not options[option]:
			print "Required option '%s' is not set! (add a --help option for help)" % option
			sys.exit()

	if options['verbose']:
		sys.stderr.write("options: " + str(options) + '\n')
	# return all options
	return options



main()
